﻿using System;
using System.Collections.Generic;
using System.Diagnostics;
using System.Threading;
using System.Threading.Tasks;

namespace Efficiency
{
    class Program
    {

        static int lowLimit = 0;
        static int hiLimit = 0;
        static EventWaitHandle ready = new AutoResetEvent(true);
        static EventWaitHandle go = new AutoResetEvent(false);
        static void Server(int packetSize, int packetCount)
        {
           
            for (int i = 0; i < packetCount; i++)
            {
                ready.WaitOne(); // wait for client ready to read next data 
                lowLimit = hiLimit;
                hiLimit += packetSize;
                go.Set(); // tell client that data are ready
            }
            ready.WaitOne(); // ensure last values was read
            lowLimit = hiLimit; //it's a sign to end
            go.Set();
            Console.WriteLine("All packages are served. Server finished");
        }
        
        static int Client(object param)
        {
            int clientNo = (int)param;
            int primesCount = 0;
            int min = 0;
            int max = 0;
            int m = 0;

            Console.WriteLine($"Client No {clientNo} is ready");
            while (true) //endless
            {
                go.WaitOne(); // wait for data
                min = lowLimit; //remember global var to local 
                max = hiLimit;
                ready.Set(); // ok we read data now, server can change lowLimit and hiLimit
                if (min == max) // this is the end
                {
                    Console.WriteLine($"Client {clientNo} has nothing to do.");
                    go.Set(); // let another client in
                    return primesCount; 
                }
                #if DEBUG
                Console.WriteLine($"Client No {clientNo} start conting primes between {min} and {max}");
                #endif
                for (int n = min; n < max; n++)
                {
                    bool isPrime = true;
                    m = n / 2;
                    for (int i = 2; i <= m; i++)
                    {
                        if (n % i == 0)
                        {
                            isPrime = false;
                            break;
                        }
                    }
                    if (isPrime == true)
                        primesCount++;
                }
                #if DEBUG
                Console.WriteLine($"Client No {clientNo} Found {primesCount} primes between {min} and {max}");
                #endif

            }
        }

        static void Main(string[] args)
        {
            //To start more tasks simultanously 8 clients and 1 server
            ThreadPool.SetMinThreads(9,0); 
        
            long sum = 0;
            int tests = 3;
            for (int j = 0; j < tests; j++) //lets do some tries to average test's results
            {
                lowLimit = 0;   
                hiLimit = 0;
                Stopwatch sw = new Stopwatch();
                //you can try different combination size and count, but size*count should be the same for example 1000000
                int size = 125000; //package size 1,5,125,125000
                int count = 8; //how many packages 1000000,200000,8000,8
                List<Task> clientTasks = new List<Task>(8);
                sw.Start();
                Task serverTask = Task.Factory.StartNew((o) =>
                {
                    var paramsArr = (object[])o; // typecast back to array of object
                    Server((int)paramsArr[0], (int)paramsArr[1]); //inside task call method Server
                }, new object[] { size, count }); //pass object as parameter to the task
                for (int i = 0; i < 8; i++)
                {
                    clientTasks.Add(new Task((parameter) => Client(parameter),i));
                }
                clientTasks.ForEach(t => t.Start()); ;
                serverTask.Wait();
                clientTasks.ForEach(t => t.Wait());
                sw.Stop();
                Console.WriteLine($"Counting last {sw.ElapsedMilliseconds} ms");
                sum += sw.ElapsedMilliseconds;
            }
            Console.WriteLine($"avr: {sum / tests}");
        }
    }
}
